"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CheckpointManager = void 0;
const assert_1 = __importDefault(require("assert"));
const common_utils_1 = require("@fluidframework/common-utils");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
class CheckpointManager {
    constructor(id, consumer) {
        this.id = id;
        this.consumer = consumer;
        this.checkpointing = false;
        this.closed = false;
    }
    /**
     * Requests a checkpoint at the given offset
     */
    async checkpoint(queuedMessage) {
        // Checkpoint calls should always be of increasing or equal value
        // Exit early if already requested checkpoint for a higher offset
        if (this.lastCheckpoint && queuedMessage.offset < this.lastCheckpoint.offset) {
            server_services_telemetry_1.Lumberjack.info("Skipping checkpoint since a request for checkpointing a higher offset has already been made", {
                lastCheckpointOffset: this.lastCheckpoint.offset,
                queuedMessageOffset: queuedMessage.offset,
                lastCheckpointPartition: this.lastCheckpoint.partition,
                queuedMessagePartition: queuedMessage.partition,
            });
            return;
        }
        // Exit early if the manager has been closed
        if (this.closed) {
            return;
        }
        // No recovery once entering an error state
        if (this.error) {
            throw this.error;
        }
        // Exit early if already caught up
        if (this.commitedCheckpoint === queuedMessage) {
            return;
        }
        // Track the highest requested offset
        this.lastCheckpoint = queuedMessage;
        // If already checkpointing allow the operation to complete to trigger another round.
        if (this.checkpointing) {
            // Create a promise that will resolve to the next checkpoint that will include the requested offset
            // and then return this as the result of checkpoint
            if (!this.pendingCheckpoint) {
                this.pendingCheckpoint = new common_utils_1.Deferred();
            }
            return this.pendingCheckpoint.promise;
        }
        // Finally begin checkpointing the offsets.
        this.checkpointing = true;
        const commitP = this.consumer.commitCheckpoint(this.id, queuedMessage);
        return commitP.then(() => {
            this.commitedCheckpoint = queuedMessage;
            this.checkpointing = false;
            // Trigger another checkpoint round if the offset has moved since the checkpoint finished and
            // resolve any pending checkpoints to it.
            if (this.lastCheckpoint && this.lastCheckpoint !== this.commitedCheckpoint) {
                (0, assert_1.default)(this.pendingCheckpoint, "Differing offsets will always result in pendingCheckpoint");
                const nextCheckpointP = this.checkpoint(this.lastCheckpoint);
                this.pendingCheckpoint.resolve(nextCheckpointP);
                this.pendingCheckpoint = undefined;
            }
            else if (this.pendingCheckpoint) {
                this.pendingCheckpoint.resolve();
                this.pendingCheckpoint = undefined;
            }
        }, 
        // eslint-disable-next-line @typescript-eslint/promise-function-async
        (error) => {
            if (error.name === "PendingCommitError") {
                server_services_telemetry_1.Lumberjack.info(`Skipping checkpoint since ${error.message}`, {
                    queuedMessageOffset: queuedMessage.offset,
                    queuedMessagePartition: queuedMessage.partition,
                });
                this.checkpointing = false;
                return;
            }
            // Enter an error state on any other commit error
            this.error = error;
            if (this.pendingCheckpoint) {
                this.pendingCheckpoint.reject(this.error);
            }
            throw error;
        });
    }
    /**
     * Checkpoints at the last received offset.
     */
    async flush() {
        if (this.lastCheckpoint) {
            server_services_telemetry_1.Lumberjack.info(`Checkpointing last recieved offset: ${this.lastCheckpoint.offset}`, {
                offset: this.lastCheckpoint.offset,
                partition: this.lastCheckpoint.partition,
            });
            return this.checkpoint(this.lastCheckpoint);
        }
    }
    /**
     * Closes the checkpoint manager - this will stop it from performing any future checkpoints
     */
    close() {
        this.closed = true;
    }
}
exports.CheckpointManager = CheckpointManager;
//# sourceMappingURL=checkpointManager.js.map